/* ggisyna - program to display sound graphically
   Copyright (C) 1997       Paul Francis Harrison
   Copyright (C) 1997-1998  Marcus Sundberg

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <fcntl.h>
#include <ggi/ggi.h>



#include "syna.h"

#define SYMBOLCOUNT 5
#define SYMBOLSPACING 40
#define SYMBOLOFFSET 60

static int mouseX = 100, mouseY = 100;
static int fadeCount = 0;
static int track=1, trackCount=1, trackStep=1;
static int symbol=-1, playing=0, showTracks=0;
static ggi_visual_t vis;
static const ggi_directbuffer *dbuf = NULL;
static int ggiusedb=0;
static uint32 *fb;
void (*gfxUpdate)(void);

int stride;
unsigned char *backbuf;
int paused=0, bpp=0;
ggi_pixel  palette[256];



static int
synachkgm(int width,int height,ggi_graphtype gt)
{
	ggi_mode mode;

#if 0    
#ifdef FAST_VIDMEM
	if (ggiCheckGraphMode(vis, width, height, GGI_AUTO, height*3, gt,
			      &mode) == 0
	    && ggiSetMode(vis, &mode) == 0
	    && (dbuf = ggiDBGetBuffer(vis, 0)) != NULL
	    && (dbuf->type & GGI_DB_SIMPLE_PLB)
	    && (dbuf->buffer.plb.stride%bpp) == 0) {
		stride = dbuf->buffer.plb.stride;
		ggiusedb = 2;
		fb = (uint32*)dbuf->write;
		backbuf = ((unsigned char*)dbuf->write)+stride*height;
		return 0;
	} else
#endif
#endif
	if (ggiCheckGraphMode(vis, width, height, GGI_AUTO, GGI_AUTO,
				     gt, &mode) == 0
	    && ggiSetMode(vis, &mode) == 0) {
		if ((dbuf = ggiDBGetBuffer(vis, 0)) != NULL
		    && (dbuf->type & GGI_DB_SIMPLE_PLB)
		    && (dbuf->buffer.plb.stride%bpp) == 0) {
			stride = dbuf->buffer.plb.stride;
			ggiusedb = 1;
		} else {
			stride = width*bpp;
			fb = (uint32*)malloc(width*height*bpp);
		}
		backbuf = malloc(width*height*bpp*2);
		return 0;
	}
	return 1;
}


static void
do8bit(void)
{
#define HIGH4MASK 0x0f0f0f0f
#define LOW4MASK  0xf0f0f0f0
	uint32 *src = (uint32*)backbuf;
	int pixels = width*height/sizeof(uint32);

	if (ggiusedb) {
		if (ggiResourceAcquire(dbuf->resource,GGI_ACTYPE_WRITE) != 0) {
			ggiPanic("Unable to acquire DirectBuffer!\n");
		}
		fb = (uint32*)dbuf->write;
	}
	if (stride == width) {
		int i = 0;
		do {
			fb[i] = (src[i]&LOW4MASK)
				| ((src[i+pixels]>>4)&HIGH4MASK);
		} while (++i != pixels);
	} else {
		int x, y;
		uint32 *dst = fb;
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				dst[x] = (src[x]&LOW4MASK)
					| ((src[x+pixels]>>4)&HIGH4MASK);
			}
			src += width/4;
			dst = (uint32*) (((uint8 *)dst)+stride);
		}
	}

	if (ggiusedb) {
		ggiResourceRelease(dbuf->resource);
	} else {
		ggiPutBox(vis, 0, 0, width, height, fb);
	}
	ggiFlush(vis);
}


static void
do16bit(void)
{
	uint8 *src = (uint8*)backbuf;
	int pixels = width*height;

	if (ggiusedb) {
		if (ggiResourceAcquire(dbuf->resource,GGI_ACTYPE_WRITE) != 0) {
			ggiPanic("Unable to acquire DirectBuffer!\n");
		}
		fb = (uint32*)dbuf->write;
	}

	if (stride == width*2) {
		int i = 0;
		int pixels2 = pixels/2;
		do {
			fb[i] =
				palette[(src[i*2]&0xf0)
				       |((src[i*2+pixels]>>4)&0x0f)]
				| palette[(src[i*2+1]&0xf0)
					 |((src[i*2+1+pixels]>>4)&0x0f)]<<16;
		} while (++i != pixels2);
	} else {
		int x, y;
		uint32 *dst = fb;
		int width2 = width/2;
		for (y = 0; y < height; y++) {
			for (x = 0; x < width2; x++) {
				dst[x] =
					palette[(src[x*2]&0xf0)
					       |((src[x*2+pixels]>>4)&0x0f)]
					| palette[(src[x*2+1]&0xf0)
						 |((src[x*2+1+pixels]>>4)&0x0f)]<<16;
			}
			src += width;
			dst = (uint32*) (((uint8 *)dst)+stride);
		}
	}

	if (ggiusedb) {
		ggiResourceRelease(dbuf->resource);
	} else {
		ggiPutBox(vis, 0, 0, width, height, fb);
	}
	ggiFlush(vis);
}


static void
do24bit(void)
{
#define seti24(addr,idx,val) \
	(*((unsigned int*)((unsigned char*)addr+(idx)*3)) = \
	 ((*((unsigned int*)((unsigned char*)addr+(idx)*3))&0xff000000) \
  | ((unsigned int)(val))))
	uint8 *src = (uint8*)backbuf;
	int pixels = width*height;

	if (ggiusedb) {
		if (ggiResourceAcquire(dbuf->resource,GGI_ACTYPE_WRITE) != 0) {
			ggiPanic("Unable to acquire DirectBuffer!\n");
		}
		fb = (uint32*)dbuf->write;
	}

	if (stride == width*3) {
		int i = 0;
		do {
			seti24(fb,i,(palette[(src[i]&0xf0)|((src[i+pixels]>>4)&0x0f)]));
		} while (++i != pixels);
	} else {
		int x, y;
		uint8 *dst = (uint8 *)fb;
		int diff = stride - width*3;
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
#ifdef READWRITE_IS_FAST
				seti24(dst,x,
				       (palette[(src[x]&0xf0)|((src[x+pixels]
								>>4)&0x0f)]));
#else
				register unsigned int pix 
					= (palette[(src[x]&0xf0)
						  |((src[x+pixels]>>4)&0x0f)]);
				*(dst++) = pix & 0xff;
				*(dst++) = (pix >> 8) & 0xff;
				*(dst++) = (pix >> 16) & 0xff;
#endif
			}
			src += width;
			dst += diff;
		}
	}

	if (ggiusedb) {
		ggiResourceRelease(dbuf->resource);
	} else {
		ggiPutBox(vis, 0, 0, width, height, fb);
	}
	ggiFlush(vis);
}

static void
do32bit(void)
{
	uint8 *src = (uint8*)backbuf;
	int pixels = width*height;

	if (ggiusedb) {
		if (ggiResourceAcquire(dbuf->resource,GGI_ACTYPE_WRITE) != 0) {
			ggiPanic("Unable to acquire DirectBuffer!\n");
		}
		fb = (uint32*)dbuf->write;
	}

	if (stride == width*4) {
		int i = 0;
		do {
			fb[i] = palette[(src[i]&0xf0)|((src[i+pixels]>>4)&0x0f)];
		} while (++i != pixels);
	} else {
		int x, y;
		uint32 *dst = fb;
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				dst[x] = palette[(src[x]&0xf0)
						| ((src[x+pixels]>>4)&0x0f)];
			}
			src += width;
			dst = (uint32*) (((uint8 *)dst)+stride);
		}
	}

	if (ggiusedb) {
		ggiResourceRelease(dbuf->resource);
	} else {
		ggiPutBox(vis, 0, 0, width, height, fb);
	}
	ggiFlush(vis);
}


void
gfxOpen(void)
{
	if (ggiInit() != 0) {
		error("starting LibGGI");
	}
	if ((vis = ggiOpen(NULL)) == NULL) {
		ggiExit();
		error("opening default visual");
	}	

	bpp=1;
	gfxUpdate = do8bit;
	if (synachkgm(width,height,GT_8BIT)
	    && (bpp=2) && (gfxUpdate = do16bit)
	    && synachkgm(width,height,GT_15BIT)
	    && synachkgm(width,height,GT_16BIT)
	    && (bpp=3) && (gfxUpdate = do24bit)
	    && synachkgm(width,height,GT_24BIT)
	    && (bpp=4) && (gfxUpdate = do32bit)
	    && synachkgm(width,height,GT_32BIT)) {
		ggiExit();
		error("setting graphics mode");
	}
	if (ggiusedb > 0)
		fprintf(stderr, "Using DirectBuffer\n");
	ggiSetFlags(vis, GGIFLAG_ASYNC);
}


void
gfxClose(void)
{
    ggiClose(vis);
    if (ggiusedb<2) {
        free(backbuf);
        if (!ggiusedb)
            free(fb);
    }
    ggiExit();
}


void
setPalette(int i,int r,int g,int b) {
    ggi_color col;
    col.r=r<<10;
    col.g=g<<10;
    col.b=b<<10;
    if (bpp==1) {
        ggiSetPalette(vis,i,1,&col);
        palette[i]=i;
    }
    else {
        palette[i]=ggiMapColor(vis,&col);
    }
}


int
processUserInput(void) {
    static int shiftstate=0;
    int statusCheckCountdown = 1000;
    int quitrequested=0;
    ggi_event ev;
    struct timeval tv = { 0, 0 };
    int mousemoved=0;
    int buttons=0;
    
    while (ggiEventPoll(vis, emKey | emPointer, &tv) != 0) {
        ggiEventRead(vis, &ev, emKey | emPointer);
        if (ev.any.type == evPtrAbsolute) {
            int oldMouseX = mouseX;
            int oldMouseY = mouseY;
            mouseX = ev.pmove.x;
            mouseY = ev.pmove.y;
            if (mouseX != oldMouseX || mouseY != oldMouseY)
                mousemoved=1;
        } else if (ev.any.type == evPtrRelative) {
            mouseX += ev.pmove.x;
            mouseY += ev.pmove.y;
            mousemoved=1;
        } else if (ev.any.type == evPtrButtonPress) {
            buttons = 1;
        } else if (ev.any.type == evKeyPress) {
            if (ev.key.label == GIIUC_Escape)
                quitrequested=1;
            else switch(ev.key.label) {
            case GIIK_PageUp:
                if (shiftstate)
                    mixUp(SYNSND_MAIN);
                else 
                    mixUp(source);
                break;
            case GIIK_PageDown:
                if (shiftstate)
                    mixDown(SYNSND_MAIN);
                else 
                    mixDown(source);
                break;
            case GIIK_ShiftL:
            case GIIK_ShiftR:
                shiftstate=1;
                break;
            default:
                break;
            }
        } else if (ev.any.type == evKeyRelease) {
            switch(ev.key.label) {
            case GIIK_ShiftL:
            case GIIK_ShiftR:
                shiftstate=0;
                break;
            default:
                break;
            }
        }
        tv.tv_sec = 0;
        tv.tv_usec = 0;
    }
    

    if (mouseX < 0) mouseX = 0; else if (mouseX > width-12) mouseX = width-12;
    if (mouseY < 0) mouseY = 0; else if (mouseY > height-12) mouseY = height-12;
	
    if (mousemoved) {
        symbol = (mouseX-SYMBOLOFFSET)/SYMBOLSPACING;
        if (mouseY > 32 || mouseX < SYMBOLOFFSET) symbol = -1;
        fadeCount = 200;
	    
        if (showTracks) {
            if (mouseX > 40)
                showTracks = 0;
        } else if ((source & SYNSND_CD)
                   && mouseX <= 40 && mouseY <= 32) {
            showTracks = 1;
            trackCount = cdGetTrackCount();
            if (trackCount)
                trackStep = (200-65)/trackCount;
            else
                showTracks = 0;
        }
    } else 
        if (fadeCount > 0) fadeCount--;
	
    if ((source & SYNSND_CD)
        && ++statusCheckCountdown > (fadeCount ? 100 : 1000)) {
	    statusCheckCountdown = 0;
	    cdGetStatus(&track,&paused,&playing);
	    
	    if (!paused && !playing) {
		    /* Loop when finished, auto play on start */
		    cdPlay(1); /* Allows you to leave it running
				  unattended (eg as a demo) */
		    cdGetStatus(&track,&paused,&playing);
		    if (!playing) fadeCount = 10;
	    }
    }
	
    if (buttons) {
        if (showTracks) {
            int which = (mouseY-32)/trackStep+1;
            if (which > 0 && which <= trackCount) {
                cdPlay(which);
                showTracks = 0;
            }
        }
	    
        fadeCount = 20;
	    
        if (source & SYNSND_CD) {
            cdGetStatus(&track,&paused,&playing);
        }
        switch(symbol) {
        case 0:
		if (source & SYNSND_CD) {
			cdPlay(track>1?track-1:1);
		}
		break;
        case 1:
		if (source & SYNSND_CD) {
			if (playing) {
				cdPause();
			} else {
				if (paused) cdResume();
				else cdPlay(1);
			}
		}
		break;
        case 2:
		if (source & SYNSND_CD) {
			cdStop();
		}
		break;
        case 3:
		if (source & SYNSND_CD) {
			cdPlay(track<1?1:track+1);
		}
		break;
        case 4: 
		quitrequested=1;
		break;
        }
        if (source & SYNSND_CD) {
		cdGetStatus(&track,&paused,&playing);
        }
    }
    
    /* Stop if quit was requested */
    if (quitrequested) {
        return 1;
    }
    return 0;
}

void
showOutput(void) {
    int symbols[SYMBOLCOUNT] = {12, 10, 13, 11, 14}; 
    unsigned int i=SYMBOLCOUNT-1;

    if (fadeCount) {
        if (source & SYNSND_CD) {
            int x,y,numberColor=0,number=track;
            if (showTracks) {
                for(i=0;i<trackCount;i++)
                    for(y=32+i*trackStep;y<32+(i*2+1)*trackStep/2+1;y++)
                        for(x=10;x<40;x++) {
                            int o = x+y*width,
                                c = (127+(int)(x+15+(sin(x/4.0+y/16.0)*5))
				     %20*(128/20)
				     +127+(int)(x+0.5+15+(sin(
					     (x+0.5)/4.0+y/16.0)*5))
				     %20*(128/20)) >> 1;
                            if (mouseY > 32 && (mouseY-32)/trackStep == i) {
				    backbuf[o+1] = c;
                            } else {
				    backbuf[o] = c;
				    if (track == i+1) backbuf[o+1] = c;
                            }
                        }
                i = (mouseY-32)/trackStep+1;
                if (mouseY > 32 && i > 0 && i <= trackCount) {
                    number = i;
                    numberColor = 1;
                }
            }
            if (number >= 10) putIcon(0,0,number/10%10,numberColor);
            putIcon(20,0,number%10,numberColor);
            
            for(i=0;i<SYMBOLCOUNT-1;i++) {
                putIcon(SYMBOLOFFSET+i*SYMBOLSPACING,0,symbols[i],
			(symbol==i?1:0));
	    }
            if (playing) {
		    putIcon(SYMBOLOFFSET+1*SYMBOLSPACING,0,symbols[1],1);
            } else {
		    putIcon(SYMBOLOFFSET+2*SYMBOLSPACING,0,symbols[2],1);
	    }
        }
        putIcon(SYMBOLOFFSET+i*SYMBOLSPACING,0,symbols[i],(symbol==i?1:0));
        putIcon(mouseX,mouseY,15,0);
        putIcon(mouseX,mouseY,16,1);
    }
    
    gfxUpdate();
}
